from typing import Any, Optional

from langchain.document_loaders import DirectoryLoader
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain.vectorstores import Pinecone


class PineconeDB:
    """Tool that queries vector database."""

    name = "pinecone"
    description = (
        "Useful for when you need to answer any questions"
        "Input should be a fully formed question."
    )

    def __init__(self, embeddings: Any, index_name: str, namespace: str):
        self.embeddings = embeddings
        self.index_name = index_name
        self.namespace = namespace
        self.docsearch = Pinecone.from_existing_index(
            self.index_name, self.embeddings, namespace=self.namespace
        )

    def _run(
        self,
        query: str,
    ) -> str:
        similar_docs = self.docsearch.similarity_search_with_score(query, k=2)
        return str(similar_docs)

    async def _arun(self, query: str):
        return NotImplementedError("pinecone async not implemented")


class LoadDocs:
    def __init__(
        self, embeddings, directory, index_name, namespace: Optional[str] = None
    ):
        self.directory = directory
        self.embeddings = embeddings
        self.index_name = index_name
        self.namespace = namespace

    def load_dir(self, directory):
        loader = DirectoryLoader(directory)
        documents = loader.load()
        return documents

    def split_docs(self, chunk_size=1000, chunk_overlap=20):
        documents = self.load_dir(self.directory)
        text_splitter = RecursiveCharacterTextSplitter(
            chunk_size=chunk_size, chunk_overlap=chunk_overlap
        )
        docs = text_splitter.split_documents(documents)
        return docs

    def docs_load(self):
        index = Pinecone.from_documents(
            self.split_docs(),
            self.embeddings,
            index_name=self.index_name,
            namespace=self.namespace,
        )
        return index



import datetime as dt
import os
import os.path
import time
from datetime import timedelta
from typing import Dict
from typing import List

from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import build
from googleapiclient.errors import HttpError

from sayvai_tools.utils.google.mail import EmailSender

# checking


class GCalendar:
    def __init__(self, scope: str, summary=None, email=None) -> None:
        """Initializes the GCalender class"""
        self.service = None
        self.creds = None
        self.SCOPE = scope
        self.calendar_id = "primary"
        self.organizer_email = email
        self.summary = summary
        if os.path.exists("token.json"):
            self.creds = Credentials.from_authorized_user_file("token.json", self.SCOPE)
        else:
            self.get_credentials()
        # if self.creds and self.creds.expired and self.creds.refresh_token:
        #     self.creds.refresh(Request())
        # os.remove('token.json')

    def get_credentials(self):
        """Gets the credentials for the user"""
        flow = InstalledAppFlow.from_client_secrets_file("credentials.json", self.SCOPE)
        self.creds = flow.run_local_server(port=0)

        with open("token.json", "w") as token:
            token.write(self.creds.to_json())

        return "Credentials obtained"

    def get_service(self):
        """Gets the service for the user"""
        try:
            self.service = build("calendar", "v3", credentials=self.creds)
            return "Service obtained"
        except Exception as e:
            print(e)
            return "Error occurred"

    def display_events(self, date, max_results: int = 10):
        """
        Displays the events for the given date
        :param date: The date for which the events are to be displayed
        :param max_results: The maximum number of events to be displayed
        :return: creates a generator with values start time,end time,summary,description,event id of the events for the given date
        """
        try:
            self.get_service()
            now = dt.datetime.combine(date, dt.time.min).isoformat() + "Z"
            end = dt.datetime.combine(date, dt.time.max).isoformat() + "Z"
            event_result = (
                self.service.events()
                .list(
                    calendarId=self.calendar_id,
                    timeMin=now,
                    timeMax=end,
                    maxResults=max_results,
                    singleEvents=True,
                    orderBy="startTime",
                )
                .execute()
            )
            events = event_result.get("items", [])
            if not events:
                return "No upcoming events found."
            for event in events:
                event_id = event["id"]
                summary = event["summary"]
                description = event["description"]
                start = event["start"].get("dateTime", event["start"].get("date"))
                end = event["end"].get("dateTime", event["end"].get("date"))

                # print(start, end)
                yield start, end, summary, description, event_id

        except Exception as e:
            print(e)
            return "Error occurred"

    def create_event(self, event: Dict):
        """
        Creates an event for the given event
        :param event: The event to be created
        :return: The event created
        """
        try:
            self.get_service()
            event = (
                self.service.events()
                .insert(calendarId=self.calendar_id, body=event)
                .execute()
            )
            # print(event)
            return "Event created", event["id"]
        except HttpError as e:
            print(e)
            return "Error occurred"

    def delete_event(self, event_id: str):
        """
        Deletes the event with the given event_id
        :param event_id:
        :return:
        """
        try:
            self.get_service()
            self.service.events().delete(
                calendarId=self.calendar_id, eventId=event_id
            ).execute()
            return "Event deleted"
        except HttpError as e:
            print(e)
            return "Error occured"

    @staticmethod
    def parse_date(input_str):
        """
        Parses the date from the input string
        :param input_str:
        :return:
        """
        try:
            year, month, day, hour, minute = map(int, input_str.split(","))
            date = dt.datetime(year, month, day, hour, minute)
            return date
        except ValueError:
            return None

    @staticmethod
    def check_is_slot_available(start_time, end_time, booked_slots):
        """
        Checks if the slot is available for the given date
        :param start_time:
        :param end_time:
        :param booked_slots:
        :return:
        """
        # Check if the time interval is between 9 AM and 5 PM

        for slot in booked_slots:
            slot_start_str, slot_end_str = slot.split(" ")
            slot_start = dt.datetime.fromisoformat(slot_start_str)
            slot_end = dt.datetime.fromisoformat(slot_end_str)

            # Check if the time interval overlaps with any booked slot
            if start_time < slot_end and end_time > slot_start:
                return False  # Slot is not available

        return True  # Slot is available

    def book_slots(
        self,
        date,
        MINIMUM_TIME_SLOT: int = 15,
        MAXIMUM_TIME_SLOT: int = 60,
        OPEN_TIME: int = 9,
        CLOSE_TIME: int = 17,
    ):
        """
        Books the slot for the given date
        checks if the provided time is in the past
        checks if the slot is already booked
        checks if the slot is within open and close time
        checks if the slot is between 15 minutes and 1 hour
        :param CLOSE_TIME:
        :param MAXIMUM_TIME_SLOT:
        :param OPEN_TIME:
        :param MINIMUM_TIME_SLOT:
        :param date:
        :return: appointment event creation
        """
        startt = time.time()

        # splits the input string into start time, end time and email
        input_pairs = date.split("/")
        start_time = self.parse_date(input_pairs[0])
        end_time = self.parse_date(input_pairs[1])
        phone = input_pairs[2]
        name = input_pairs[3]

        current_datetime = dt.datetime.now()

        # Check if the provided date and time are in the past
        if start_time < current_datetime:
            return "The provided date and time are in the past."

        specific_date = start_time.date()  # Use the date from the input

        booked_slots = []

        # calls display_events function to get the booked slots for the given date
        for start, end, summary, descript, event_id in self.display_events(
            specific_date
        ):
            booked_slots.append(start + " " + end)
            if summary == "day is not available for booking":

                # formatting the start, end and start_time, end_time to compare
                start = dt.datetime.strptime(start, "%Y-%m-%dT%H:%M:%S%z")
                end = dt.datetime.strptime(end, "%Y-%m-%dT%H:%M:%S%z")
                start_time_with_timezone = start_time.replace(tzinfo=start.tzinfo)
                end_time_with_timezone = end_time.replace(tzinfo=end.tzinfo)

                # print(start_time_with_timezone, end_time_with_timezone)
                # print(start, end)
                # checks if the given time interval overlaps with the blocked day time interval
                if (
                    start < start_time_with_timezone < end
                    and start < end_time_with_timezone < end
                ) or (
                    start_time_with_timezone < end and end_time_with_timezone > start
                ):
                    # print(start, end)
                    return descript

        time_interval = (
            start_time.isoformat() + "+05:30" + " " + end_time.isoformat() + "+05:30"
        )

        start_time, end_time = time_interval.split(" ")
        start_time = dt.datetime.fromisoformat(start_time)
        end_time = dt.datetime.fromisoformat(end_time)

        # checks if the input time interval is valid
        if end_time <= start_time:
            return "End time should be greater than start time."

        # checks if the given time interval is between minimum and maximum time slot
        duration = end_time - start_time
        if duration < dt.timedelta(
            minutes=MINIMUM_TIME_SLOT
        ) or duration > dt.timedelta(minutes=MAXIMUM_TIME_SLOT):
            return "The slot should be between 15 minutes and 1 hour."

        # checks if the given time interval is within the open and close time
        if (
            OPEN_TIME <= start_time.hour < CLOSE_TIME
            and OPEN_TIME <= end_time.hour < CLOSE_TIME
        ):
            # Check if the slot is available
            if self.check_is_slot_available(start_time, end_time, booked_slots):
                self.summary = self.summary.replace("{name}", name).replace(
                    "{phone}", phone
                )
                events = {
                    "summary": self.summary,
                    "location": "Coimbatore, Tamil Nadu, India",
                    "description": "default description",
                    "start": {
                        "dateTime": start_time.isoformat(),
                        "timeZone": "IST",
                    },
                    "end": {
                        "dateTime": end_time.isoformat(),
                        "timeZone": "IST",
                    },
                    "recurrence": ["RRULE:FREQ=DAILY;COUNT=1"],
                    "attendees": [
                        {"email": self.organizer_email},
                    ],
                }
                endd = time.time()
                total = endd - startt
                return self.create_event(events)
            else:
                return "The slot is already booked.", booked_slots

        else:
            return "The slot is not within 9 AM - 5 PM."

    # def update_event(self, date: str):
    #     self.get_service()
    #     now = dt.datetime.combine(date, dt.time.min).isoformat() + 'Z'
    #     event_result = self.service.events().list(calendarId=self.calendar_id, timeMin=now,
    #                                               maxResults=20, singleEvents=True,
    #                                               orderBy='startTime').execute()
    #     events = event_result.get('items', [])
    #     print(events)

    def block_day(
        self,
        date: str,
        contacts: List[str],
        organizer: str,
        smtp_username: str,
        smtp_password: str,
    ):
        """
        Blocks the day for the given date and time
        cancels the appointments that are in the given time interval and send a mail to the user
        :param contacts:
        :param organizer:
        :param smtp_username:
        :param smtp_password:
        :param date:
        :return: block event creation
        organizer : Role [doctor, dentist, event holder]
        """
        input_pairs = date.split("/")
        start_time = self.parse_date(input_pairs[0])
        end_time = self.parse_date(input_pairs[1])

        events = {
            "summary": "day is not available for booking",
            "description": f"the doctor is not available from {start_time} to {end_time}",
            "start": {
                "dateTime": start_time.isoformat(),
                "timeZone": "IST",
            },
            "end": {
                "dateTime": end_time.isoformat(),
                "timeZone": "IST",
            },
            "attendees": [
                {"email": self.organizer_email},
            ],
        }

        specific_date = start_time.date()  # Use the date from the input

        # booked_slots_block_day = []

        # calls display_events function to get the booked slots for the given date
        for start, end, summary, descript, event_id in self.display_events(
            specific_date
        ):
            start = dt.datetime.strptime(start, "%Y-%m-%dT%H:%M:%S%z")
            end = dt.datetime.strptime(end, "%Y-%m-%dT%H:%M:%S%z")
            start_time_with_timezone = start_time.replace(tzinfo=start.tzinfo)
            end_time_with_timezone = end_time.replace(tzinfo=end.tzinfo)

            # checks if there are any appointments in the given time interval of the block day and deletes the appointment
            if (
                (
                    (
                        start < start_time_with_timezone < end
                        and start < end_time_with_timezone < end
                    )
                    or (
                        start_time_with_timezone < end
                        and end_time_with_timezone > start
                    )
                )
            ) and summary != "day is not available for booking":
                # time = start.isoformat() + ' ' + end.isoformat()
                # booked_slots_block_day.append((time, event_id))

                self.delete_event(event_id)
                # event_id = event_id.split('_')[0]

                # TODO: send message to the user that the slot is deleted either via whatsapp

                # send mail to the user that the appointment is cancelled
                mail_class = EmailSender(
                    organizer_email=self.organizer_email,
                    smtp_username=smtp_username,
                    smtp_password=smtp_password,
                )
                for email in contacts:
                    mail_class.send_email(
                        receiver_email=email,
                        subject="cancelling the appointment",
                        message=f"sorry for the inconvenience caused. the {organizer} is not available on the "
                        f"given date:{specific_date}  and time from {start_time} to {end_time}."
                        f" please book another slot.",
                    )

        # print(booked_slots_block_day)
        # creates the block dat event
        return self.create_event(events)

    def free_slots(self, date):
        """
        Finds the free slots for the given date
        :param date:
        :return: list containing the free slots
        """

        input_pairs = date.split("/")
        working_start = self.parse_date(input_pairs[0])
        working_end = self.parse_date(input_pairs[1])

        specific_date = working_start.date()
        booked_slots = []

        # calls display_events function to get the booked slots for the given date
        for start, end, summary, descript, event_id in self.display_events(
            specific_date
        ):
            booked_slots.append(start + " " + end)

        working_start = working_start.isoformat()
        working_end = working_end.isoformat()

        working_start = dt.datetime.strptime(
            f"{working_start}+05:30", "%Y-%m-%dT%H:%M:%S%z"
        )
        working_end = dt.datetime.strptime(
            f"{working_end}+05:30", "%Y-%m-%dT%H:%M:%S%z"
        )
        booked_slots = []

        specific_date = working_start.date()

        # calls display_events function to get the booked slots for the given date
        for start, end, summary, descript, event_id in self.display_events(
            specific_date
        ):
            booked_slots.append(start + " " + end)

        booked_ranges = []
        for slot in booked_slots:
            start_str, end_str = slot.split()
            start_time = dt.datetime.fromisoformat(start_str)
            end_time = dt.datetime.fromisoformat(end_str)
            booked_ranges.append((start_time, end_time))

        # Initialize the list of available time slots
        available_slots = []

        # Initialize the current time as the start of the working hours
        current_time = working_start

        # Iterate through the working hours
        while current_time < working_end:
            # Find the next available time slot
            next_slot_start = current_time
            while current_time < working_end:
                overlapping = False
                for booked_range in booked_ranges:
                    if (
                        current_time < booked_range[1]
                        and current_time + timedelta(minutes=30) > booked_range[0]
                    ):
                        overlapping = True
                        break
                if overlapping:
                    break
                current_time += timedelta(minutes=30)
            next_slot_end = current_time

            if next_slot_start < next_slot_end:
                available_slots.append(
                    f"{next_slot_start.strftime('%H:%M')} - {next_slot_end.strftime('%H:%M')} free"
                )

            # Move to the next 30-minute slot
            current_time += timedelta(minutes=30)

        free = []
        # Print the available time slots
        for slot in available_slots:
            free.append(slot)

        return ("booked_slots:", booked_slots), ("free slots:", free)



import pandas as pd
from sqlalchemy import create_engine, text

from sayvai_tools.utils.google.sheets import GSheets


class SQLSheet:
    name = "sqlsheet"
    description = (
        "Useful for exporting SQL query results to a spreadsheet."
        "You can use this to generate reports."
    )

    def __init__(self, uri: str, sheet_path: str = "output.xlsx"):
        self.connection = create_engine(uri).connect()
        self.sheet_path = sheet_path

    def _run(self, query: str):
        df = pd.read_sql_query(text(query), self.connection)
        df.to_excel(self.sheet_path, index=False)
        return f"Data has been exported to {self.sheet_path}"


class SQLGSheet:
    name = "sqlgsheet"
    description = (
        "Useful for exporting SQL query results to a G-Sheets."
        "You can use this to generate reports."
    )

    def __init__(self, uri: str):
        self.gs = GSheets()
        self.connection = create_engine(uri).connect()

    def _run(self, query: str):
        df = pd.read_sql_query(text(query), self.connection)
        data_dict = df.to_dict("split")
        columns = data_dict["columns"]
        result = [columns]  # Start with the header
        result.extend(data_dict["data"])  # Add the data
        self.gs.create_sheet()
        self.gs.update_values(result)
        return "Data has been exported to Google Sheets"



"""Google Sheets API""" ""
import json
import gspread
import datetime
from gspread import utils


class GSheets:
    """Google Sheets API""" ""

    def __init__(self) -> None:
        self.gc, _user = self.authenticate()
        pass

    def get_credentials(self):
        with open("credentials.json") as f:
            data = json.load(f)
        return data

    def authenticate(self):
        credentials = self.get_credentials()
        # Check Creditentials is valid
        if not credentials:
            raise Exception("Invalid Credentials")
        # Authorize API
        return gspread.oauth_from_dict(credentials)

    def get_date(self):
        now = datetime.datetime.now()
        return now.strftime("%Y-%m-%d %H:%M:%S")

    def create_sheet(self):
        # # Authorize API
        self.sheet = self.gc.create(f"Records-Upto-{self.get_date()}")

    def update_values(self, values: list):
        if not self.sheet:
            self.create_sheet()
        worksheet = self.sheet.get_worksheet(0)
        start_cell = "A1"
        end_cell = utils.rowcol_to_a1(len(values), len(values[0]))

        # Combine the start and end cells to form the range
        range_to_update = f"{start_cell}:{end_cell}"

        # Update the range with the values
        worksheet.update(
            range_to_update, values=values, value_input_option="USER_ENTERED"
        )

    def get_values(self):
        # For now, we will only work with the first sheet
        worksheet = self.sheet.get_worksheet(0)
        values = worksheet.get_all_values()
        return values



import yaml
from google.cloud import speech_v1p1beta1 as speech

from sayvai_tools.utils.voice.recording import record


class STT:
    def __init__(
        self,
        model: str = "default",
        sample_rate: int = 44100,
        lang_code: str = "en-IN",
        automatic_punctuation: bool = True,
        enhanced: bool = True,
        second_lang_codes: list[str] = ["en-US"],
        no_of_audio_channels: int = 2,
        sep_rec_per_channel: bool = True,
        spoken_punctuation: bool = True,
        word_confidence: bool = True,
        spoken_emoji: bool = True,
        max_alt: int = 0,
        audio_format: str = "default",
        interact_type: str = "default",
        naics_code: int | None = None,
        mic_distance: str = "default",
        media_type: str = "default",
        record_device_type: str = "default",
        record_device_name: str | None = None,
        audio_topic: str = "general",
        speech_context_path: str | None = None,
    ):
        self.model = model
        self.sample_rate_hertz = sample_rate
        self.language_code = lang_code
        self.enable_automatic_punctuation = automatic_punctuation
        self.use_enhanced = enhanced
        self.alternate_language_codes = second_lang_codes
        self.audio_channel_count = no_of_audio_channels
        self.enable_separate_recognition_per_channel = sep_rec_per_channel
        self.enable_spoken_punctuation = spoken_punctuation
        self.enable_spoken_emojis = spoken_emoji
        self.max_alternatives = max_alt
        self.audio_format = audio_format
        self.enable_word_confidence = word_confidence
        self.interaction_type = interact_type
        self.naics_code = naics_code
        self.microphone_distance = mic_distance
        self.original_media_type = media_type
        self.recording_device_type = record_device_type
        self.recording_device_name = record_device_name
        self.audio_topic = audio_topic
        self.speech_context_path = speech_context_path
        self.state = True

    def audio_encoding(self):
        if self.audio_format == "mp3":
            return speech.RecognitionConfig.AudioEncoding.MP3
        elif self.audio_format == "flac":
            return speech.RecognitionConfig.AudioEncoding.FLAC
        elif self.audio_format == "l16":
            return speech.RecognitionConfig.AudioEncoding.LINEAR16
        elif self.audio_format == "mulaw":
            return speech.RecognitionConfig.AudioEncoding.MULAW
        elif self.audio_format == "amr":
            return speech.RecognitionConfig.AudioEncoding.AMR
        elif self.audio_format == "amr-wb":
            return speech.RecognitionConfig.AudioEncoding.AMR_WB
        elif self.audio_format == "ogg-opus":
            return speech.RecognitionConfig.AudioEncoding.OGG_OPUS
        elif self.audio_format == "speex":
            return speech.RecognitionConfig.AudioEncoding.SPEEX_WITH_HEADER_BYTE
        elif self.audio_format == "webm-opus":
            return speech.RecognitionConfig.AudioEncoding.WEBM_OPUS
        else:
            return speech.RecognitionConfig.AudioEncoding.ENCODING_UNSPECIFIED

    def create_reg_config(self):

        config_mp3 = speech.RecognitionConfig(
            model=self.model,
            encoding=self.audio_encoding(),
            sample_rate_hertz=self.sample_rate_hertz,
            enable_automatic_punctuation=self.enable_automatic_punctuation,
            language_code=self.language_code,
            use_enhanced=self.use_enhanced,
            alternative_language_codes=self.alternate_language_codes,
            audio_channel_count=self.audio_channel_count,
            enable_separate_recognition_per_channel=self.enable_separate_recognition_per_channel,
            enable_spoken_punctuation=self.enable_spoken_punctuation,
            enable_spoken_emojis=self.enable_spoken_emojis,
            max_alternatives=self.max_alternatives,
            speech_contexts=self.speech_context(),
            enable_word_confidence=self.enable_word_confidence,
            metadata=self.recognition_meta_data(),
        )
        return config_mp3

    def read_audio(self):
        byte_data = record()
        try:
            audio = speech.RecognitionAudio(content=byte_data)
            return audio
        except:
            return self.read_audio()

    def generate_text(self):
        self.check_for_bounds()
        if self.state == True:
            try:
                speech_client = speech.SpeechClient()
                audio = self.read_audio()
                try:
                    response_standard_mp3 = speech_client.recognize(
                        config=self.create_reg_config(), audio=audio
                    )
                    return response_standard_mp3.results[0].alternatives[0].transcript

                except:
                    response_standard_mp3 = speech_client.long_running_recognize(
                        config=self.create_reg_config(), audio=audio
                    )
                    return (
                        response_standard_mp3.result()
                        .results[0]
                        .alternatives[0]
                        .transcript
                    )
            except:
                return self.generate_text()

        else:
            raise ValueError("Speech Parameters Invalid")

    @staticmethod
    def load_phrase_set(file_path: str):
        with open(file_path, "r", encoding="utf-8") as yaml_file:
            loaded_data = yaml.safe_load(yaml_file)
            return loaded_data

    def speech_context(self):
        if self.speech_context_path is not None:
            speech_context = self.load_phrase_set(self.speech_context_path)
            return speech_context
        else:
            return None

    def check_speech_context(self):
        speech_context = self.speech_context()
        if speech_context is not None:
            for i in speech_context:
                if i.get("boost") > 20 or i.get("boost") < 0:
                    self.state = False

    def recognition_meta_data(self):
        metadata = speech.RecognitionMetadata(
            interaction_type=self.select_interaction_type(),
            industry_naics_code_of_audio=self.naics_code,
            microphone_distance=self.select_microphone_distance(),
            original_media_type=self.select_media_type(),
            recording_device_name=self.recording_device_name,
            recording_device_type=self.select_recording_device_type(),
            audio_topic=self.audio_topic,
        )
        return metadata

    def select_interaction_type(self):
        if self.interaction_type == "discussion":
            return speech.RecognitionMetadata.InteractionType.DISCUSSION
        elif self.interaction_type == "presentation":
            return speech.RecognitionMetadata.InteractionType.PRESENTATION
        elif self.interaction_type == "phone_call":
            return speech.RecognitionMetadata.InteractionType.PHONE_CALL
        elif self.interaction_type == "voice_mail":
            return speech.RecognitionMetadata.InteractionType.VOICEMAIL
        elif self.interaction_type == "professionally_produced":
            return speech.RecognitionMetadata.InteractionType.PROFESSIONALLY_PRODUCED
        elif self.interaction_type == "voice_search":
            return speech.RecognitionMetadata.InteractionType.VOICE_SEARCH
        elif self.interaction_type == "voice_command":
            return speech.RecognitionMetadata.InteractionType.VOICE_COMMAND
        elif self.interaction_type == "dictation":
            return speech.RecognitionMetadata.InteractionType.DICTATION
        else:
            return (
                speech.RecognitionMetadata.InteractionType.INTERACTION_TYPE_UNSPECIFIED
            )

    def select_microphone_distance(self):
        if self.microphone_distance == "nearfield":
            return speech.RecognitionMetadata.MicrophoneDistance.NEARFIELD
        elif self.microphone_distance == "midfield":
            return speech.RecognitionMetadata.MicrophoneDistance.MIDFIELD
        elif self.microphone_distance == "farfield":
            return speech.RecognitionMetadata.MicrophoneDistance.FARFIELD
        else:
            return (
                speech.RecognitionMetadata.MicrophoneDistance.MICROPHONE_DISTANCE_UNSPECIFIED
            )

    def select_media_type(self):
        if self.original_media_type == "audio":
            return speech.RecognitionMetadata.OriginalMediaType.AUDIO
        elif self.original_media_type == "video":
            return speech.RecognitionMetadata.OriginalMediaType.VIDEO
        else:
            return (
                speech.RecognitionMetadata.OriginalMediaType.ORIGINAL_MEDIA_TYPE_UNSPECIFIED
            )

    def select_recording_device_type(self):
        if self.recording_device_type == "smartphone":
            return speech.RecognitionMetadata.RecordingDeviceType.SMARTPHONE
        elif self.recording_device_type == "pc":
            return speech.RecognitionMetadata.RecordingDeviceType.PC
        elif self.recording_device_type == "phone_line":
            return speech.RecognitionMetadata.RecordingDeviceType.PHONE_LINE
        elif self.recording_device_type == "vehicle":
            return speech.RecognitionMetadata.RecordingDeviceType.VEHICLE
        elif self.recording_device_type == "other_outdoor_device":
            return speech.RecognitionMetadata.RecordingDeviceType.OTHER_OUTDOOR_DEVICE
        elif self.recording_device_type == "other_indoor_device":
            return speech.RecognitionMetadata.RecordingDeviceType.OTHER_INDOOR_DEVICE
        else:
            return (
                speech.RecognitionMetadata.RecordingDeviceType.RECORDING_DEVICE_TYPE_UNSPECIFIED
            )

    def check_for_bounds(self):
        self.check_speech_context()



"""Google Forms"""
from __future__ import print_function
from apiclient import discovery
from httplib2 import Http
from typing import Any, Optional, List, Text
from oauth2client import client, file, tools

SCOPES = "https://www.googleapis.com/auth/forms.body"
DISCOVERY_DOC = "https://forms.googleapis.com/$discovery/rest?version=v1"


class GForms:
    """Google Forms"""
    def __init__(self, formTitle: str = "Test Form"):
        store = file.Storage('token.json')
        creds = None
        if not creds or creds.invalid:
            flow = client.flow_from_clientsecrets('client_secrets.json', SCOPES)
            creds = tools.run_flow(flow, store)

        self.form_service = discovery.build('forms', 'v1', http=creds.authorize(
            Http()), discoveryServiceUrl=DISCOVERY_DOC, static_discovery=False)
        
        self.formTitle = formTitle
        self.formID = self.create_form(title=formTitle)
        self.__update_form()
        pass
    
    def create_form(self, title: str = "Test Form"):
        NEW_FORM = {
            "info": {
                "title": title,
            },
        }
        result = self.form_service.forms().create(body=NEW_FORM).execute()
        return result["formId"]
    
    def __update_form(self):
        UPDATE_FORM = {
            "requests": [
                {
                    "updateFormInfo": {
                        "info": {
                            "title": self.formTitle,
                            "description": "created by sayvai-io",
                            "documentTitle": "sayvai-io",
                        },
                        "updateMask": "*",
                    }
                }
            ]
        }
        self.form_service.forms().batchUpdate(formId= self.formID, body=UPDATE_FORM).execute()
    
    def add_question(self, question: str, location : int  = 0) -> str:
        # Creates the initial form
        # Adds the question to the form 
        NEW_QUESTION = {
            "includeFormInResponse": True,
            "requests": [
                {
                    "createItem": {
                        "item": {
                            "itemId": str(location),
                            "title": question,
                            "description": "Please fill in the blank",
                            "questionItem": {
                                "question" : {
                                    "required": True,
                                    "textQuestion": {
                                        "paragraph": False,
                                    }
                                }
                            }
                        },
                        "location": {
                            "index": location,
                        }
                    }
                    
                }
            ]
        }
        self.form_service.forms().batchUpdate(formId=self.formID, body=NEW_QUESTION).execute()
        return "Question has been added"
    
    
    def add_multiple_questions(self, questions: List[str]) -> str:
        for index, question in enumerate(questions):
            self.add_question(question=question, location=index)
        return "All Questions have been added"
    
    

    def get_form(self, form_title: str):
        # Prints the result to show the question has been added
        # get form id from title 
        raise NotImplementedError("This method is not implemented yet")
    
    
    def get_form_url(self, form_title: str):
        raise NotImplementedError("This method is not implemented yet")


"""Conversational Human """
import os
from typing import Callable, Optional

from elevenlabs import play
from langchain.callbacks.manager import CallbackManagerForToolRun
from langchain.pydantic_v1 import Field

from sayvai_tools.utils.voice.stt import STT
from sayvai_tools.utils.voice.tts import ElevenlabsAudioStreaming


class ConversationalHuman:
    """Tool that asks user for input."""

    name: str = "human"
    description: str = (
        "You can ask a human for guidance when you think you "
        "got stuck or you are not sure what to do next. "
        "The input should be a question for the human."
    )
    # prompt_func: Callable[[str], None] = Field(default_factory=lambda: _print_func)
    # input_func: Callable = Field(default_factory=lambda: input)

    def __init__(self, api_key: str, g_api_key: str, phrase_set_path: str) -> None:
        self.stt = STT(audio_format="mp3", speech_context_path=phrase_set_path)
        self.tts = ElevenlabsAudioStreaming(api_key=api_key)
        self.g_api_key = g_api_key

    def _run(
        self,
        query: str,
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> str:
        """Use the Human input tool."""
        os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = self.g_api_key
        inputbytes = self.tts.audio_streaming(
            query,
            model="eleven_multilingual_v1",
            voice="Adam",
            audio_streaming=True,
            stability=0.5,
            similarity=0.5,
            # api_key= self.api_key
        )
        play(inputbytes)

        # self.prompt_func(query)
        # return self.input_func()
        return self.stt.generate_text()



from sayvai_tools.utils.google.gcalendar import GCalendar


class Block:
    name = "Block Day"
    description = "Block a day in the calendar"

    def __init__(
        self, organizer: str, smtp_username: str, smtp_password: str, scope: str
    ):
        self.organizer = organizer
        self.smtp_username = smtp_username
        self.smtp_password = smtp_password
        self.scope = scope
        self.summary = None
        self.email = None
        self.cal = GCalendar(scope=self.scope, email=self.email, summary=self.summary)

    def _run(self, date: str, contacts: list):
        self.cal.block_day(
            date=date,
            contacts=contacts,
            organizer=self.organizer,
            smtp_username=self.smtp_username,
            smtp_password=self.smtp_password,
        )
        return "Day is blocked"

    def _arun(self, date: str, contacts: list):
        raise NotImplementedError("This method is not implemented yet")



from sayvai_tools.utils.database.dbbase import SQLDatabase
from sayvai_tools.utils.database.dbchain import SQLDatabaseChain
from sayvai_tools.utils.google.gcalendar import GCalendar
from sayvai_tools.utils.google.mail import EmailSender
from sayvai_tools.utils.google.sheets import GSheets
from sayvai_tools.utils.google.forms import GForms
from sayvai_tools.utils.voice.stt import STT
from sayvai_tools.utils.voice.tts import ElevenlabsAudioStreaming
from sayvai_tools.utils.microsoft.excel import Excel

__all__ = [
    "GCalendar",
    "SQLDatabase",
    "SQLDatabaseChain",
    "ElevenlabsAudioStreaming",
    "EmailSender",
    "GSheets",
    "STT",
    "GForms",
    "Excel",
]



from typing import Optional

from langchain.schema import BasePromptTemplate
from langchain.schema.language_model import BaseLanguageModel
from sqlalchemy.engine import Engine

from sayvai_tools.tools.sql_database.prompt import PROMPT, SQL_PROMPTS
from sayvai_tools.utils.database.dbbase import SQLDatabase
from sayvai_tools.utils.database.dbchain import SQLDatabaseChain


class Database:
    """Tool that queries vector database."""

    name = "Database"
    description = (
        "Useful for when you need to access sql database"
        "Input should be a natural language"
    )

    def __init__(
        self,
        llm: BaseLanguageModel,
        engine: Engine,
        prompt: Optional[BasePromptTemplate] = None,
        verbose: bool = False,
        k: int = 5,
    ):
        self.llm = llm
        self.engine = engine
        self.prompt = prompt
        self.verbose = verbose
        self.k = k

    def _run(self, query: str) -> str:
        db = SQLDatabase(engine=self.engine)

        if self.prompt is not None:
            prompt_to_use = self.prompt
        elif db.dialect in SQL_PROMPTS:
            prompt_to_use = SQL_PROMPTS[db.dialect]
        else:
            prompt_to_use = PROMPT
        inputs = {
            "input": lambda x: x["question"] + "\nSQLQuery: ",
            "top_k": lambda _: self.k,
            "table_info": lambda x: db.get_table_info(
                table_names=x.get("table_names_to_use")
            ),
        }
        if "dialect" in prompt_to_use.input_variables:
            inputs["dialect"] = lambda _: (db.dialect, prompt_to_use)

        sql_db_chain = SQLDatabaseChain.from_llm(
            llm=self.llm, db=db, prompt=prompt_to_use, verbose=self.verbose
        )

        return sql_db_chain.run(query)

    async def _arun(self, query: str):

        raise NotImplementedError("SQL database async not implemented")



import datetime as dt

from sqlalchemy import text

from sayvai_tools.utils.google.gcalendar import GCalendar


class RetrievePhone:
    def __init__(self, pool, scope: str):
        self.pool = pool
        self.cursor = self.pool.connect()
        self.scope = scope
        self.summary = None
        self.email = None
        self.cal = GCalendar(scope=self.scope, email=self.email, summary=self.summary)

    name = "Retrieve Email"
    description = "Retrieve Email from the calendar"

    def _run(self, date: str):
        input_dates = date.split("/")
        start_time = self.cal.parse_date(input_dates[0])
        end_time = self.cal.parse_date(input_dates[1])

        specific_date = start_time.date()

        email_list = []

        for start, end, summary, descript, event_id in self.cal.display_events(
            specific_date
        ):
            start = dt.datetime.strptime(start, "%Y-%m-%dT%H:%M:%S%z")
            end = dt.datetime.strptime(end, "%Y-%m-%dT%H:%M:%S%z")
            start_time_with_timezone = start_time.replace(tzinfo=start.tzinfo)
            end_time_with_timezone = end_time.replace(tzinfo=end.tzinfo)

            # checks if there are any appointments in the given time interval of the block day and deletes the appointment
            if (
                (
                    (
                        start < start_time_with_timezone < end
                        and start < end_time_with_timezone < end
                    )
                    or (
                        start_time_with_timezone < end
                        and end_time_with_timezone > start
                    )
                )
            ) and summary != "day is not available for booking":
                event_id = event_id.split("_")[0]

                query = self.cursor.execute(
                    text(
                        f"""SELECT phone FROM patient_info WHERE event_id = '{event_id}';"""
                    )
                )
                email = query.fetchone()[0]
                email_list.append(email)

        return email_list

    def _arun(self):
        raise NotImplementedError("This method is not implemented yet")



import datetime


class GetDate:
    def _run(self, tool_input: str) -> str:
        """Use the tool."""
        current_time = datetime.datetime.now()
        formatted_time = current_time.strftime("%A, %B %d, %Y %I:%M %p")
        return formatted_time

    async def _arun(self, tool_input: str) -> str:
        """Use the tool asynchronously."""
        raise ("Not NotImplemented Error")
        # return None



"""TTS file for sayvai_tools."""
from elevenlabs import generate, save, stream, voices
from elevenlabs.api.voice import Voice
from elevenlabs.simple import VOICES_CACHE, is_voice_id

# elevenlabs.set_api_key("431f452112cab175b80762e50e525c8f")


# from elevenlabs_audio_streaming.constant import VALID_MODELS


class ElevenlabsAudioStreaming:
    def __init__(self, api_key) -> None:
        self.api_key = api_key

    @staticmethod
    def check_voice(voice, stability, similarity):
        """
        checks if the voice is available
        Returns: Voice class containing voice_id, name, setting

        """
        if isinstance(voice, str):
            voice_str = voice
        if is_voice_id(voice):
            voice = Voice(voice_id=voice)
        else:
            # Check if voice is in cache
            voice = next((v for v in VOICES_CACHE if v.name == voice_str), None)
            # if the voice not in VOICE_CACHE, call the api to check is voice is available
            if not voice:
                voice = (
                    next((v for v in voices() if v.name == voice_str), None)
                    if not voice
                    else voice
                )
        # if voice not found raise ValueError
        if not voice:
            raise ValueError(f"Voice '{voice_str}' not found.")

        voice.settings.stability = stability
        voice.settings.similarity_boost = similarity

    @staticmethod
    def check_model(model):
        """
        checks if the model is available is not
        Returns: value error if model not found

        """

    def audio_streaming(
        self, text, voice, model, audio_streaming, stability, similarity
    ):
        api_key = self.api_key
        """
        passes the text to elevenlabs api to play the audio

        """
        if not isinstance(audio_streaming, bool):
            raise ValueError(f"Invalid Streaming '{audio_streaming}' value.")

        if not isinstance(text, str):
            raise ValueError("Text input must be a string")

        self.check_voice(voice, stability, similarity)
        self.check_model(model)

        # if Audio streaming is true
        if audio_streaming:
            audio_stream = generate(
                text=text,
                stream=audio_streaming,
                voice=voice,
                model=model,
                api_key=api_key,
            )
            # audio_stream is a generator with byte values that cannot be saved directly using save function
            # so we add all the byte values in generator to a single variable and save the audio
            byte_values = bytearray()
            for byte_chunk in audio_stream:
                byte_values += byte_chunk

            # Convert the accumulated byte values to a bytes object
            final_byte_data = bytes(byte_values)
            return final_byte_data
            # save(final_byte_data, "E:/Text-to-speech/src/audio buffer/audio.wav")
            # play(final_byte_data)

        # if Audio streaming is False
        else:
            audio = generate(text=text, voice=voice, model=model, api_key=api_key)
            return audio
            # save(audio, "E:/Text-to-speech/src/audio buffer/audio.wav")
            # play(audio)

    @staticmethod
    def avail_voices():
        """
        Returns: returns voice names and corresponding labels

        """
        voice = voices()

        for voice in voice:
            print(voice.name, voice.labels)



import io

import numpy as np
import pyaudio
from pydub import AudioSegment
from pydub.silence import split_on_silence

# Constants for audio input
FORMAT = pyaudio.paInt16
CHANNELS = 1
RATE = 44100
CHUNK_SIZE = 16000

# Initialize PyAudio
audio = pyaudio.PyAudio()


def record():
    def match_target_amplitude(aChunk, target_dBFS):
        """Normalize given audio chunk"""
        change_in_dBFS = target_dBFS - aChunk.dBFS
        return aChunk.apply_gain(change_in_dBFS)

    # Open a stream to capture audio from the microphone
    stream = audio.open(
        format=FORMAT,
        channels=CHANNELS,
        rate=RATE,
        input=True,
        output=False,
        frames_per_buffer=CHUNK_SIZE,
    )

    # Create a list to store audio chunks
    audio_chunks = []

    # Recording flag
    recording = False

    # Record audio until interrupted
    try:
        j = 0
        while True:
            data = stream.read(CHUNK_SIZE)
            audio_data = np.frombuffer(data, dtype=np.int16)

            # Check if audio data is silence
            is_silence = np.max(audio_data) < 350
            # print(np.max(audio_data))

            if recording:
                if is_silence:
                    # End of an audio chunk
                    recording = False
                    if len(audio_chunks) > 0:
                        # Process the recorded audio chunk
                        song = AudioSegment(
                            data=b"".join(audio_chunks),
                            sample_width=2,
                            frame_rate=RATE,
                            channels=CHANNELS,
                        )

                        # Split the chunk on silence
                        chunks = split_on_silence(
                            song, min_silence_len=1500, silence_thresh=-35
                        )
                        silence_chunk = AudioSegment.silent(duration=500)
                        combined_chunks = AudioSegment.empty()
                        for i in chunks:
                            combined_chunks += i

                        # Create a silence chunk that's 0.5 seconds (500 ms) long for padding
                        audio_chunk = silence_chunk + combined_chunks + silence_chunk
                        # Normalize the entire chunk
                        normalized_chunk = match_target_amplitude(audio_chunk, -20.0)

                        # Export the audio chunk with a new bitrate
                        mp3_buffer = io.BytesIO()
                        normalized_chunk.export(
                            mp3_buffer, format="mp3", bitrate="192k"
                        )

                        mp3_bytes = mp3_buffer.getvalue()
                        mp3_buffer.close()
                        return mp3_bytes

                        # Clear the audio chunks list
                        audio_chunks.clear()

                        break
                else:
                    # Continue recording audio data
                    audio_chunks.append(data)

            else:
                # Start recording when non-silent audio is detected
                if not is_silence:
                    recording = True
                    audio_chunks.append(data)

        else:
            stream.stop_stream()
            stream.close()
            audio.terminate()
            audio.close(stream=stream)

    except KeyboardInterrupt:
        print("Recording stopped.")



# tool for forms
from typing import List, Dict, Any
from sayvai_tools.utils.google.forms import GForms


class FormTool:
    def __init__(self, scope: str, title: str = "Test Form"):
        self.scope = scope
        self.form = GForms(formTitle=title)

    name = "Form Tool"
    description = "Tool for forms"

    def _run(self, questions: List[str]):
        self.form.add_multiple_questions(questions=questions)
        return "Questions have been added"

    def _arun(self):
        pass



from sayvai_tools.utils.google.sheets import GSheets


class Sheets:

    name = "sheets"
    description = (
        "Useful for creating and updating sheets."
        "You can use sheets to generate reports."
        """Input: [["Name", "Age"], ["John", "20"], ["Jane", "19"]]"""
    )

    def __init__(self):
        self.gs = GSheets()

    def _run(self, values: list):
        self.gs.create_sheet()
        self.gs.update_values(values)
        return "Data has been exported to Google Sheets"

    async def _arun(self, date: str):
        raise NotImplementedError("Sheets async not implemented")



from sayvai_tools.utils.google.gcalendar import GCalendar


class Calendar:
    name = "calendar"
    description = "You can ask calendar tool to create an event for you."

    def __init__(self, scope: str):
        self.scope = scope
        self.cal = GCalendar(self.scope)

    def _run(self, date: str):
        return self.cal.book_slots(date)

    async def _arun(self, date: str):
        raise NotImplementedError("Calendar async not implemented")



"""base tool for IO"""
from typing import Callable

from elevenlabs import play
from langchain.pydantic_v1 import Field
from sayvai_tools.utils.voice.tts import ElevenlabsAudioStreaming


class VoiceOutputRun:
    """Tool that asks user for input."""

    name: str = "voice"
    description: str = (
        "You can ask a human for guidance when you think you "
        "got stuck or you are not sure what to do next. "
        "The input should be a question for the human."
    )

    def __init__(self, api_key: str) -> None:
        self.api_key = api_key
        assert isinstance(self.api_key, str)
        pass

    def _run(
        self,
        query: str,
    ):
        """Use the Human input tool."""
        # input_func: Callable = Field(default_factory=lambda: input)
        tts = ElevenlabsAudioStreaming()
        inputbytes = tts.audio_streaming(
            query,
            model="eleven_multilingual_v1",
            voice="Adam",
            audio_streaming=True,
            stability=0.5,
            similarity=0.5,
            api_key=self.api_key,
        )
        play(inputbytes)



import smtplib
import ssl
from email.message import EmailMessage


class EmailSender:
    def __init__(self, organizer_email, smtp_username, smtp_password):
        self.organizer_email = organizer_email
        self.smtp_username = smtp_username
        self.smtp_password = smtp_password
        self.smtp_server = "smtp.gmail.com"
        self.smtp_port = 465  # SSL port

    def send_email(self, receiver_email, subject, message):
        em = EmailMessage()
        em.set_content(message)
        em["Subject"] = subject
        em["From"] = self.organizer_email
        em["To"] = receiver_email

        context = ssl.create_default_context()
        try:
            with smtplib.SMTP_SSL(
                self.smtp_server, self.smtp_port, context=context
            ) as server:
                server.login(self.smtp_username, self.smtp_password)
                server.sendmail(self.organizer_email, receiver_email, em.as_string())
                print("Email sent successfully")
        except Exception as e:
            print(e)
            
    def send_multiple_email(self, receiver_emails, subject, message):
        for receiver_email in receiver_emails:
            self.send_email(receiver_email, subject, message)
        



## Read Local CSV File
import os
import pandas as pd


class Excel:
    def __init__(self, path):
        self.path = path

    def read(self):
        if os.path.exists(self.path):
            return pd.read_excel(self.path)
        else:
            print("File not found")
            return None

    def write(self, data, sheet_name="Sheet1"):
        if os.path.exists(self.path):
            writer = pd.ExcelWriter(self.path, engine='xlsxwriter')
            data.to_excel(writer, sheet_name=sheet_name)
            writer.save()
            writer.close()
        else:
            print("File not found")
            return None
    
    def read_column(self, column_name):
        df = self.read()
        return df[column_name].tolist()


from sayvai_tools.utils.google.gcalendar import GCalendar
from sqlalchemy import text


class CalendarSql:
    name = "calendar_sql"
    description = "You can ask calendar_sql tool to create an event for you."

    def __init__(self, pool, scope: str, email: str, summary: str):
        self.pool = pool
        self.cursor = self.pool.connect()
        self.scope = scope
        self.email = email
        self.summary = summary
        self.cal = GCalendar(self.scope, email=self.email, summary=self.summary)

    def _run(self, details: str):
        start_time, end_time, phone, name = details.split("/")
        start_time = self.cal.parse_date(start_time)
        end_time = self.cal.parse_date(end_time)

        specific_date = start_time.date()
        date_string = specific_date.strftime("%Y-%m-%d")

        query = self.cursor.execute(text(f"""SELECT phone_number FROM patient_info;"""))
        phone_number = query.fetchall()
        phone_number = [i[0] for i in phone_number]
        result = self.cal.book_slots(details)
        if len(result) == 2 and result[0] == "Event created":
            msg, event_id = result[0], result[1]
            if details.split("/")[2] not in phone_number:
                query = self.cursor.execute(
                    text(
                        f"""INSERT INTO patient_info (name, phone_number, start_time, end_time, event_id, appointment_date) VALUES ('{name}', '{phone}'
                                            , '{start_time}', '{end_time}','{event_id}','{date_string}');"""
                    )
                )
                # query.commit()
            else:
                query = self.cursor.execute(
                    text(
                        f"""UPDATE patient_info SET start_time = '{start_time}', end_time = '{end_time}', event_id = '{event_id}', appointment_date = '{date_string}' WHERE phone_number = '{phone}';"""
                    )
                )
                # query.commit()

            return msg
        else:
            return result

    async def _arun(self, date: str):
        raise NotImplementedError("Calendar async not implemented")



# Author: Aathikabilan
import asyncio
import os

from sayvai_tools.utils import EmailSender, Excel


class SendMail:
    name = "send_mail"
    description = "You can ask send_mail tool to send mail for you."

    def __init__(self, path: str):
        SMTP_USERNAME = os.getenv("SMTP_USERNAME")
        SMTP_PASSWORD = os.getenv("SMTP_PASSWORD")
        ORGANIZER_EMAIL = os.getenv("ORGANIZER_EMAIL")
        self.excel = Excel(path=path)
        self.emails = self.excel.read_column("email")
        self.email_sender = EmailSender(ORGANIZER_EMAIL, SMTP_USERNAME, SMTP_PASSWORD)

    def _run(self, content: str):
        self.email_sender.send_multiple_email(self.emails, "Sayvai", content)
        return "Mail sent successfully"

    async def _arun(self, content: str):
        raise NotImplementedError("Send mail via async not implemented")


from setuptools import find_packages, setup

core_requirements = [
    "langchain",
    "langchain_experimental",
    "google-api-python-client",
    "openai",
    "tiktoken",
    "SQLAlchemy",
    "elevenlabs==0.2.24",
    "google_auth_oauthlib",
    "google-auth-httplib2",
    "numpy",
    "google",
    "google-cloud-speech",
    "cloud-sql-python-connector",
    "pg8000",
    "pinecone-client",
    "pgvector",
    "pandas",
    "openpyxl",
    "gspread==5.11.3",
    "pyaudio",
    "pydub",
    "oauth2client"
]

setup(
    name="sayvai-tools",
    version="0.0.1",
    description="Tools for the assistant",
    author="sayvai-io",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    install_requires=core_requirements,
    extras_require={
        "dev": [
            "pytest",
            "pylint",
        ],
    },
    zip_safe=False,
)



from sayvai_tools.utils.google.gcalendar import GCalendar


class DisplayEvents:
    name = "Display Events"
    description = "Display events from the calendar"

    def __init__(self, scope: str):
        self.scope = scope
        self.summary = None
        self.email = None
        self.cal = GCalendar(scope=self.scope, email=self.email, summary=self.summary)

    def _run(self, date: str):
        date = self.cal.parse_date(date)
        specific_date = date.date()
        booked_slots = []
        for start, end, summary, descript, event_id in self.cal.display_events(
            specific_date
        ):
            booked_slots.append((start + " " + end, summary))

        return booked_slots

    def _arun(self):
        raise NotImplementedError("This method is not implemented yet")



from typing import Any

from langchain.document_loaders import DirectoryLoader
from langchain.text_splitter import CharacterTextSplitter
from langchain.vectorstores.pgvector import PGVector


class PGVectorDB:
    """Tool that queries vector database."""

    name = "pgvector"
    description = (
        "Useful for when you need to answer any questions"
        "Input should be a fully formed question."
    )

    def __init__(self, embeddings: Any, collection_name: str, connection_string: str):
        self.embeddings = embeddings
        self.connection_string = connection_string
        self.collection_name = collection_name
        self.docsearch = PGVector(
            collection_name=self.collection_name,
            connection_string=self.connection_string,
            embedding_function=self.embeddings,
        )

    def _run(
        self,
        query: str,
        k: int = 2,
    ) -> str:
        similar_docs = self.docsearch.similarity_search_with_score(query, k=k)
        return similar_docs

    async def _arun(self, query: str):
        return NotImplementedError("pinecone async not implemented")


class LoadDocs:
    """The PGVector Module will try to create a table with the name of the collection.
    So, make sure that the collection name is unique and the user has the permission to create a table."""

    def __init__(
        self, directory, embeddings: Any, collection_name: str, connection_string: str
    ):
        self.directory = directory
        self.embeddings = embeddings
        self.connection_string = connection_string
        self.collection_name = collection_name

    def load_dir(self, directory):
        loader = DirectoryLoader(directory)
        documents = loader.load()
        return documents

    def split_docs(self, chunk_size=1000, chunk_overlap=20):
        documents = self.load_dir(self.directory)
        text_splitter = CharacterTextSplitter(
            chunk_size=chunk_size, chunk_overlap=chunk_overlap
        )
        docs = text_splitter.split_documents(documents)
        return docs

    def docs_load(self):
        db = PGVector.from_documents(
            embedding=self.embeddings,
            documents=self.split_docs(),
            collection_name=self.collection_name,
            connection_string=self.connection_string,
        )
        return db
